# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause
import numpy as np
import pandas as pd
from numpy import nan
from numpy.testing import assert_almost_equal
from packaging.version import Version
from sklearn import __version__ as sklearn_version
from sklearn.datasets import load_iris
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LinearRegression
from sklearn.metrics import accuracy_score, make_scorer, roc_auc_score
from sklearn.model_selection import GridSearchCV, GroupKFold
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import Pipeline

from mlxtend.classifier import SoftmaxRegression
from mlxtend.data import boston_housing_data
from mlxtend.feature_selection import SequentialFeatureSelector as SFS
from mlxtend.utils import assert_raises


def nan_roc_auc_score(y_true, y_score, average="macro", sample_weight=None):
    if len(np.unique(y_true)) != 2:
        return np.nan
    else:
        return roc_auc_score(
            y_true, y_score, average=average, sample_weight=sample_weight
        )


def dict_compare_utility(d_actual, d_desired, decimal=2):
    assert d_actual.keys() == d_desired.keys(), "%s != %s" % (d_actual, d_desired)
    for i in d_actual:
        err_msg = "d_actual[%s]['feature_idx']" " != d_desired[%s]['feature_idx']" % (
            i,
            i,
        )
        assert d_actual[i]["feature_idx"] == d_desired[i]["feature_idx"], err_msg
        assert_almost_equal(
            actual=d_actual[i]["avg_score"],
            desired=d_desired[i]["avg_score"],
            decimal=decimal,
            err_msg=(
                "d_actual[%s]['avg_score']" " != d_desired[%s]['avg_score']" % (i, i)
            ),
        )
        assert_almost_equal(
            actual=d_actual[i]["cv_scores"],
            desired=d_desired[i]["cv_scores"],
            decimal=decimal,
            err_msg=(
                "d_actual[%s]['cv_scores']" " != d_desired[%s]['cv_scores']" % (i, i)
            ),
        )


def test_run_default():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()
    sfs = SFS(estimator=knn, verbose=0)
    sfs.fit(X, y)
    assert sfs.k_feature_idx_ == (3,)


def test_fit_params():
    iris = load_iris()
    X = iris.data
    y = iris.target
    sample_weight = np.ones(X.shape[0])
    forest = RandomForestClassifier(n_estimators=100, random_state=123)
    sfs = SFS(estimator=forest, verbose=0)
    sfs.fit(X, y, sample_weight=sample_weight)
    assert sfs.k_feature_idx_ == (3,)


def test_kfeatures_type_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()
    name = "k_features"
    expect = f"{name} must be between 1 and X.shape[1]."
    sfs = SFS(estimator=knn, verbose=0, k_features=0)
    assert_raises(AttributeError, expect, sfs.fit, X, y)


def test_kfeatures_type_2():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()
    expect = "k_features must be a positive integer, tuple, or string"
    sfs = SFS(estimator=knn, verbose=0, k_features=set())
    assert_raises(AttributeError, expect, sfs.fit, X, y)


def test_kfeatures_type_3():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()
    expect = "k_features tuple min value must be between 1 and X.shape[1]."
    sfs = SFS(estimator=knn, verbose=0, k_features=(0, 5))
    assert_raises(AttributeError, expect, sfs.fit, X, y)


def test_kfeatures_type_4():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()
    expect = "k_features tuple max value must be between 1 and X.shape[1]."
    sfs = SFS(estimator=knn, verbose=0, k_features=(1, 5))
    assert_raises(AttributeError, expect, sfs.fit, X, y)


def test_kfeatures_type_5():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier()
    expect = (
        "The min k_features value must be smaller" " than the max k_features value."
    )
    sfs = SFS(estimator=knn, verbose=0, k_features=(3, 1))
    assert_raises(AttributeError, expect, sfs.fit, X, y)


def test_knn_wo_cv():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(knn, k_features=3, forward=True, floating=False, cv=0, verbose=0)
    sfs1 = sfs1.fit(X, y)
    expect = {
        1: {
            "avg_score": 0.95999999999999996,
            "cv_scores": np.array([0.96]),
            "feature_idx": (3,),
        },
        2: {
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
            "feature_idx": (2, 3),
        },
        3: {
            "avg_score": 0.97333333333333338,
            "cv_scores": np.array([0.97333333]),
            "feature_idx": (1, 2, 3),
        },
    }
    dict_compare_utility(d_actual=sfs1.subsets_, d_desired=expect, decimal=2)


def test_knn_cv3():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(knn, k_features=3, forward=True, floating=False, cv=4, verbose=0)
    sfs1 = sfs1.fit(X, y)
    sfs1.subsets_
    expect = {
        1: {
            "avg_score": 0.9599928876244666,
            "cv_scores": np.array([0.974, 0.947, 0.919, 1.0]),
            "feature_idx": (3,),
        },
        2: {
            "avg_score": 0.95993589743589736,
            "cv_scores": np.array([0.974, 0.947, 0.919, 1.0]),
            "feature_idx": (2, 3),
        },
        3: {
            "avg_score": 0.9732,
            "cv_scores": np.array([0.974, 1.0, 0.946, 0.973]),
            "feature_idx": (1, 2, 3),
        },
    }

    if Version(sklearn_version) < Version("1.0"):
        expect[1]["avg_score"] = 0.95299145299145294
        expect[1]["cv_scores"] = (np.array([0.974, 0.947, 0.892, 1.0]),)

    if Version(sklearn_version) < Version("0.22"):
        expect[1]["cv_scores"] = np.array([0.97435897, 0.94871795, 0.88888889, 1.0])
        expect[2]["cv_scores"] = np.array([0.97435897, 0.94871795, 0.91666667, 1.0])
        expect[2]["avg_score"] = 0.97275641025641035
        expect[3]["cv_scores"] = np.array([0.97435897, 1.0, 0.94444444, 0.97222222])

    dict_compare_utility(d_actual=sfs1.subsets_, d_desired=expect, decimal=2)


def test_knn_cv3_groups():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(
        knn,
        k_features=3,
        forward=True,
        floating=False,
        cv=GroupKFold(n_splits=3),
        verbose=0,
    )
    np.random.seed(1630672634)
    groups = np.random.randint(0, 6, size=len(y))
    sfs1 = sfs1.fit(X, y, groups=groups)
    # print(sfs1.subsets_)
    expect = {
        1: {
            "cv_scores": np.array([0.97916667, 0.93877551, 0.96226415]),
            "feature_idx": (3,),
            "avg_score": 0.9600687759380482,
        },
        2: {
            "cv_scores": np.array([0.95833333, 0.93877551, 0.98113208]),
            "feature_idx": (1, 3),
            "avg_score": 0.9594136396697044,
        },
        3: {
            "cv_scores": np.array([0.97916667, 0.95918367, 0.94339623]),
            "feature_idx": (1, 2, 3),
            "avg_score": 0.9605821888503829,
        },
    }
    dict_compare_utility(d_actual=sfs1.subsets_, d_desired=expect, decimal=3)


def test_knn_rbf_groupkfold():
    nan_roc_auc_scorer = make_scorer(nan_roc_auc_score)
    rng = np.random.RandomState(123)
    iris = load_iris()
    X = iris.data
    # knn = KNeighborsClassifier(n_neighbors=4)
    forest = RandomForestClassifier(n_estimators=100, random_state=123)
    bool_01 = [True if item == 0 else False for item in iris["target"]]
    bool_02 = [True if (item == 1 or item == 2) else False for item in iris["target"]]
    groups = []
    y_new = []
    for ind, _ in enumerate(bool_01):
        if bool_01[ind]:
            groups.append("attribute_A")
            y_new.append(0)
        if bool_02[ind]:
            throw = rng.rand()
            if throw < 0.5:
                groups.append("attribute_B")
            else:
                groups.append("attribute_C")
            throw2 = rng.rand()
            if throw2 < 0.5:
                y_new.append(0)
            else:
                y_new.append(1)
    y_new_bool = [True if item == 1 else False for item in y_new]
    cv_obj = GroupKFold(n_splits=3)
    cv_obj_list = list(cv_obj.split(X, np.array(y_new_bool), groups))
    sfs1 = SFS(
        forest,
        k_features=3,
        forward=True,
        floating=False,
        cv=cv_obj_list,
        scoring=nan_roc_auc_scorer,
        verbose=0,
    )
    sfs1 = sfs1.fit(X, y_new)
    expect = {
        1: {
            "cv_scores": np.array([0.52, nan, 0.72]),
            "avg_score": 0.62,
            "feature_idx": (1,),
        },
        2: {
            "cv_scores": np.array([0.42, nan, 0.65]),
            "avg_score": 0.53,
            "feature_idx": (1, 2),
        },
        3: {
            "cv_scores": np.array([0.47, nan, 0.63]),
            "avg_score": 0.55,
            "feature_idx": (1, 2, 3),
        },
    }

    dict_compare_utility(d_actual=sfs1.subsets_, d_desired=expect, decimal=1)


def test_knn_option_sfs():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(knn, k_features=3, forward=True, floating=False, cv=4, verbose=0)
    sfs1 = sfs1.fit(X, y)
    assert sfs1.k_feature_idx_ == (1, 2, 3)


def test_knn_option_sffs():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs2 = SFS(knn, k_features=3, forward=True, floating=True, cv=4, verbose=0)
    sfs2 = sfs2.fit(X, y)
    assert sfs2.k_feature_idx_ == (1, 2, 3)


def test_knn_option_sbs():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs3 = SFS(knn, k_features=3, forward=False, floating=False, cv=4, verbose=0)
    sfs3 = sfs3.fit(X, y)
    assert sfs3.k_feature_idx_ == (1, 2, 3)


def test_knn_option_sfbs():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs4 = SFS(knn, k_features=3, forward=False, floating=True, cv=4, verbose=0)
    sfs4 = sfs4.fit(X, y)
    assert sfs4.k_feature_idx_ == (1, 2, 3)


def test_knn_option_sfbs_tuplerange_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=3)
    sfs4 = SFS(knn, k_features=(1, 3), forward=False, floating=True, cv=4, verbose=0)
    sfs4 = sfs4.fit(X, y)
    assert round(sfs4.k_score_, 3) == 0.967, sfs4.k_score_
    assert sfs4.k_feature_idx_ == (0, 2, 3), sfs4.k_feature_idx_


def test_knn_option_sfbs_tuplerange_2():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=3)
    sfs4 = SFS(knn, k_features=(1, 4), forward=False, floating=True, cv=4, verbose=0)
    sfs4 = sfs4.fit(X, y)
    assert round(sfs4.k_score_, 3) == 0.967, sfs4.k_score_
    assert sfs4.k_feature_idx_ == (0, 2, 3), sfs4.k_feature_idx_


def test_knn_option_sffs_tuplerange_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=3)
    sfs4 = SFS(knn, k_features=(1, 3), forward=True, floating=True, cv=4, verbose=0)
    sfs4 = sfs4.fit(X, y)
    assert round(sfs4.k_score_, 3) == 0.967, sfs4.k_score_
    assert sfs4.k_feature_idx_ == (0, 2, 3), sfs4.k_feature_idx_


def test_knn_option_sfs_tuplerange_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=3)
    sfs4 = SFS(knn, k_features=(1, 3), forward=True, floating=False, cv=4, verbose=0)
    sfs4 = sfs4.fit(X, y)
    assert round(sfs4.k_score_, 3) == 0.967, sfs4.k_score_
    assert sfs4.k_feature_idx_ == (0, 2, 3), sfs4.k_feature_idx_


def test_knn_option_sbs_tuplerange_1():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=3)
    sfs4 = SFS(knn, k_features=(1, 3), forward=False, floating=False, cv=4, verbose=0)
    sfs4 = sfs4.fit(X, y)
    assert round(sfs4.k_score_, 3) == 0.967, sfs4.k_score_
    assert sfs4.k_feature_idx_ == (0, 2, 3), sfs4.k_feature_idx_


def test_knn_scoring_metric():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs5 = SFS(knn, k_features=3, forward=False, floating=True, cv=4, verbose=0)
    sfs5 = sfs5.fit(X, y)

    if Version(sklearn_version) < Version("0.22"):
        assert round(sfs5.k_score_, 4) == 0.9728
    else:
        assert round(sfs5.k_score_, 4) == 0.9732

    sfs6 = SFS(knn, k_features=3, forward=False, floating=True, cv=4, verbose=0)
    sfs6 = sfs6.fit(X, y)
    if Version(sklearn_version) < Version("0.22"):
        assert round(sfs5.k_score_, 4) == 0.9728
    else:
        assert round(sfs5.k_score_, 4) == 0.9732

    sfs7 = SFS(
        knn, k_features=3, forward=False, floating=True, scoring="f1_macro", cv=4
    )
    sfs7 = sfs7.fit(X, y)
    if Version(sklearn_version) < Version("0.22"):
        assert round(sfs5.k_score_, 4) == 0.9727
    else:
        assert round(sfs5.k_score_, 4) == 0.9732


def test_regression():
    X, y = boston_housing_data()
    lr = LinearRegression()
    sfs_r = SFS(
        lr,
        k_features=13,
        forward=True,
        floating=False,
        scoring="neg_mean_squared_error",
        cv=10,
        verbose=0,
    )
    sfs_r = sfs_r.fit(X, y)
    assert len(sfs_r.k_feature_idx_) == 13

    if Version(sklearn_version) < Version("0.20"):
        assert round(sfs_r.k_score_, 4) == -34.7631, round(sfs_r.k_score_, 4)
    else:
        assert round(sfs_r.k_score_, 4) == -34.7053, round(sfs_r.k_score_, 4)


def test_regression_sffs():
    X, y = boston_housing_data()
    lr = LinearRegression()
    sfs_r = SFS(
        lr,
        k_features=11,
        forward=True,
        floating=True,
        scoring="neg_mean_squared_error",
        cv=10,
        verbose=0,
    )
    sfs_r = sfs_r.fit(X, y)
    assert sfs_r.k_feature_idx_ == (0, 1, 3, 4, 6, 7, 8, 9, 10, 11, 12)


def test_regression_sbfs():
    X, y = boston_housing_data()
    lr = LinearRegression()
    sfs_r = SFS(
        lr,
        k_features=3,
        forward=False,
        floating=True,
        scoring="neg_mean_squared_error",
        cv=10,
        verbose=0,
    )
    sfs_r = sfs_r.fit(X, y)
    assert sfs_r.k_feature_idx_ == (7, 10, 12), sfs_r.k_feature_idx_


def test_regression_in_range():
    X, y = boston_housing_data()
    lr = LinearRegression()
    sfs_r = SFS(
        lr,
        k_features=(1, 13),
        forward=True,
        floating=False,
        scoring="neg_mean_squared_error",
        cv=10,
        verbose=0,
    )
    sfs_r = sfs_r.fit(X, y)
    assert len(sfs_r.k_feature_idx_) == 9

    if Version(sklearn_version) < Version("0.20"):
        assert round(sfs_r.k_score_, 4) == -31.1537, round(sfs_r.k_score_, 4)
    else:
        assert round(sfs_r.k_score_, 4) == -31.1299, round(sfs_r.k_score_, 4)


def test_clone_params_fail():
    class Perceptron(object):
        def __init__(self, eta=0.1, epochs=50, random_seed=None, print_progress=0):
            self.eta = eta
            self.epochs = epochs
            self.random_seed = random_seed
            self.print_progress = print_progress
            self._is_fitted = False

        def _fit(self, X, y, init_params=True):
            self._check_target_array(y, allowed={(0, 1)})
            y_data = np.where(y == 0, -1.0, 1.0)

            if init_params:
                self.b_, self.w_ = self._init_params(
                    weights_shape=(X.shape[1], 1),
                    bias_shape=(1,),
                    random_seed=self.random_seed,
                )
                self.cost_ = []

            rgen = np.random.RandomState(self.random_seed)
            for i in range(self.epochs):
                errors = 0

                for idx in self._yield_minibatches_idx(
                    rgen=rgen, n_batches=y_data.shape[0], data_ary=y_data, shuffle=True
                ):
                    update = self.eta * (y_data[idx] - self._to_classlabels(X[idx]))
                    self.w_ += (update * X[idx]).reshape(self.w_.shape)
                    self.b_ += update
                    errors += int(update != 0.0)

                if self.print_progress:
                    self.print_progress(
                        iteration=i + 1, n_iter=self.epochs, cost=errors
                    )
                self.cost_.append(errors)
            return self

        def _net_input(self, X):
            """Net input function"""
            return (np.dot(X, self.w_) + self.b_).flatten()

        def _to_classlabels(self, X):
            return np.where(self._net_input(X) < 0.0, -1.0, 1.0)

        def _predict(self, X):
            return np.where(self._net_input(X) < 0.0, 0, 1)

    expect = (
        "Cannot clone object. You should provide an "
        "instance of scikit-learn estimator instead of a class."
    )

    assert_raises(
        TypeError,
        expect,
        SFS,
        Perceptron,
        scoring="accuracy",
        k_features=3,
        clone_estimator=True,
    )


def test_clone_params_pass():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=1)
    sfs1 = SFS(
        lr,
        k_features=2,
        forward=True,
        floating=False,
        scoring="accuracy",
        cv=0,
        clone_estimator=True,
        verbose=0,
        n_jobs=1,
    )
    sfs1 = sfs1.fit(X, y)
    assert sfs1.k_feature_idx_ == (1, 3)


def test_transform_not_fitted():
    iris = load_iris()
    X = iris.data
    knn = KNeighborsClassifier(n_neighbors=4)

    sfs1 = SFS(
        knn,
        k_features=2,
        forward=True,
        floating=False,
        cv=0,
        clone_estimator=False,
        verbose=0,
        n_jobs=1,
    )

    expect = "SequentialFeatureSelector has not been fitted, yet."

    assert_raises(AttributeError, expect, sfs1.transform, X)


def test_get_metric_dict_not_fitted():
    knn = KNeighborsClassifier(n_neighbors=4)

    sfs1 = SFS(
        knn,
        k_features=2,
        forward=True,
        floating=False,
        cv=0,
        clone_estimator=False,
        verbose=0,
        n_jobs=1,
    )

    expect = "SequentialFeatureSelector has not been fitted, yet."

    assert_raises(AttributeError, expect, sfs1.get_metric_dict)


def test_cv_generator_raises():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)

    groups = np.arange(len(y)) // 50
    cv_gen = GroupKFold(n_splits=3).split(X, y, groups)

    expect = (
        "Input cv is a generator object, which is not supported. "
        "Instead please input an iterable yielding train, test splits. "
        "This can usually be done by passing a cross-validation "
        "generator to the built-in list function. I.e. "
        "cv=list(<cv-generator>)"
    )

    assert_raises(
        TypeError, expect, SFS, knn, k_features=2, cv=cv_gen, verbose=0, n_jobs=1
    )


def test_keyboard_interrupt():
    iris = load_iris()
    X = iris.data
    y = iris.target

    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(
        knn,
        k_features=3,
        forward=True,
        floating=False,
        cv=3,
        clone_estimator=False,
        verbose=5,
        n_jobs=1,
    )

    sfs1._TESTING_INTERRUPT_MODE = True
    out = sfs1.fit(X, y)

    assert len(out.subsets_.keys()) > 0
    assert sfs1.interrupted_


def test_gridsearch():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=2)

    sfs1 = SFS(estimator=knn, k_features=3, forward=True, floating=False, cv=5)

    pipe = Pipeline([("sfs", sfs1), ("knn", knn)])

    param_grid = [
        {"sfs__k_features": [1, 2, 3, 4], "sfs__estimator__n_neighbors": [1, 2, 3, 4]}
    ]

    if Version(sklearn_version) < Version("0.24.1"):
        gs = GridSearchCV(
            estimator=pipe,
            param_grid=param_grid,
            n_jobs=1,
            iid=False,
            cv=5,
            refit=False,
        )
    else:
        gs = GridSearchCV(
            estimator=pipe, param_grid=param_grid, n_jobs=1, cv=5, refit=False
        )

    gs = gs.fit(X, y)

    assert gs.best_params_["sfs__k_features"] == 3


def test_string_scoring_clf():
    iris = load_iris()
    X = iris.data
    y = iris.target
    knn = KNeighborsClassifier(n_neighbors=4)
    sfs1 = SFS(knn, k_features=3, cv=0)
    sfs1 = sfs1.fit(X, y)

    sfs2 = SFS(knn, k_features=3, scoring="accuracy", cv=0)
    sfs2 = sfs2.fit(X, y)

    sfs3 = SFS(knn, k_features=3, scoring=make_scorer(accuracy_score), cv=0)
    sfs3 = sfs2.fit(X, y)

    assert sfs1.k_score_ == sfs2.k_score_
    assert sfs1.k_score_ == sfs3.k_score_


def test_max_feature_subset_size_in_tuple_range():
    X, y = boston_housing_data()
    lr = LinearRegression()

    sfs = SFS(
        lr,
        k_features=(1, 5),
        forward=False,
        floating=True,
        scoring="neg_mean_squared_error",
        cv=10,
    )

    sfs = sfs.fit(X, y)
    assert len(sfs.k_feature_idx_) == 5


def test_max_feature_subset_best():
    X, y = boston_housing_data()
    lr = LinearRegression()

    sfs = SFS(lr, k_features="best", forward=True, floating=False, cv=10)

    sfs = sfs.fit(X, y)
    assert sfs.k_feature_idx_ == (1, 3, 5, 7, 8, 9, 10, 11, 12)


def test_max_feature_subset_parsimonious():
    X, y = boston_housing_data()
    lr = LinearRegression()

    sfs = SFS(lr, k_features="parsimonious", forward=True, floating=False, cv=10)

    sfs = sfs.fit(X, y)
    assert sfs.k_feature_idx_ == (5, 10, 11, 12)


def test_check_pandas_dataframe_fit():
    for floating in [True, False]:
        iris = load_iris()
        X = iris.data
        y = iris.target
        lr = SoftmaxRegression(random_seed=1)
        sfs1 = SFS(
            lr,
            k_features=2,
            forward=True,
            floating=floating,
            scoring="accuracy",
            cv=0,
            verbose=0,
            n_jobs=1,
        )

        df = pd.DataFrame(
            X, columns=["sepal len", "sepal width", "petal len", "petal width"]
        )

        sfs1 = sfs1.fit(X, y)
        assert sfs1.k_feature_idx_ == (1, 3)
        assert sfs1.k_feature_names_ == ("1", "3")
        assert sfs1.subsets_[2]["feature_names"] == ("1", "3")

        sfs1 = sfs1.fit(df, y)
        assert sfs1.subsets_[1]["feature_names"] == ("petal width",)
        assert sfs1.subsets_[2]["feature_names"] == ("sepal width", "petal width")
        assert sfs1.subsets_[1]["feature_idx"] == (3,)
        assert sfs1.subsets_[2]["feature_idx"] == (1, 3)
        assert sfs1.k_feature_idx_ == (1, 3)
        assert sfs1.k_feature_names_ == ("sepal width", "petal width")

        sfs1._TESTING_INTERRUPT_MODE = True
        out = sfs1.fit(df, y)
        assert len(out.subsets_.keys()) > 0
        assert sfs1.interrupted_
        assert sfs1.subsets_[1]["feature_names"] == ("petal width",)
        assert sfs1.k_feature_idx_ == (3,)
        assert sfs1.k_feature_names_ == ("petal width",)


def test_check_pandas_dataframe_fit_backward():
    for floating in [True, False]:
        iris = load_iris()
        X = iris.data
        y = iris.target
        lr = SoftmaxRegression(random_seed=1)
        sfs1 = SFS(
            lr,
            k_features=2,
            forward=False,
            floating=floating,
            scoring="accuracy",
            cv=0,
            verbose=0,
            n_jobs=1,
        )

        df = pd.DataFrame(
            X, columns=["sepal len", "sepal width", "petal len", "petal width"]
        )

        sfs1 = sfs1.fit(X, y)
        assert sfs1.k_feature_idx_ == (1, 2)
        assert sfs1.k_feature_names_ == ("1", "2")
        assert sfs1.subsets_[2]["feature_names"] == ("1", "2")

        sfs1 = sfs1.fit(df, y)
        assert sfs1.subsets_[3]["feature_names"] == (
            "sepal len",
            "sepal width",
            "petal len",
        )
        assert sfs1.subsets_[2]["feature_names"] == ("sepal width", "petal len")
        assert sfs1.subsets_[3]["feature_idx"] == (0, 1, 2)
        assert sfs1.subsets_[2]["feature_idx"] == (1, 2)
        assert sfs1.k_feature_idx_ == (1, 2)
        assert sfs1.k_feature_names_ == ("sepal width", "petal len")

        sfs1._TESTING_INTERRUPT_MODE = True
        out = sfs1.fit(df, y)
        assert len(out.subsets_.keys()) > 0
        assert sfs1.interrupted_
        assert sfs1.subsets_[3]["feature_names"] == (
            "sepal len",
            "sepal width",
            "petal len",
        )
        assert sfs1.k_feature_idx_ == (0, 1, 2)
        assert sfs1.k_feature_names_ == ("sepal len", "sepal width", "petal len")


def test_check_pandas_dataframe_transform():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=1)
    sfs1 = SFS(
        lr,
        k_features=2,
        forward=True,
        floating=False,
        scoring="accuracy",
        cv=0,
        verbose=0,
        n_jobs=1,
    )

    df = pd.DataFrame(
        X, columns=["sepal length", "sepal width", "petal length", "petal width"]
    )
    sfs1 = sfs1.fit(df, y)
    assert sfs1.k_feature_idx_ == (1, 3)
    assert (150, 2) == sfs1.transform(df).shape


def test_invalid_estimator():
    expect = "Estimator must have an ._estimator_type for infering `scoring`"
    assert_raises(AttributeError, expect, SFS, PCA())

    class PCA2(PCA):
        def __init__(self):
            super().__init__()
            self._estimator_type = "something"

    expect = "Estimator must be a Classifier or Regressor."
    assert_raises(AttributeError, expect, SFS, PCA2())


def test_invalid_k_features():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=1)

    sfs1 = SFS(lr, k_features=(1, 2, 3), scoring="accuracy")
    expect = "k_features tuple must consist of 2 elements, a min and a max value."
    assert_raises(AttributeError, expect, sfs1.fit, X, y)

    sfs1 = SFS(lr, k_features="something", scoring="accuracy")
    expect = 'If a string argument is provided, it must be "best" or "parsimonious"'
    assert_raises(AttributeError, expect, sfs1.fit, X, y)


def test_verbose():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=1)

    sfs1 = SFS(lr, k_features=1, scoring="accuracy", verbose=1)
    sfs1.fit(X, y)


def test_check_pandas_dataframe_with_feature_groups():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=1)

    df = pd.DataFrame(
        X, columns=["sepal length", "sepal width", "petal length", "petal width"]
    )

    sfs1 = SFS(
        lr,
        k_features=2,
        forward=True,
        floating=False,
        scoring="accuracy",
        feature_groups=[
            ["sepal length", "petal length"],
            ["sepal width"],
            ["petal width"],
        ],
        cv=0,
        verbose=0,
        n_jobs=1,
    )

    sfs1 = sfs1.fit(df, y)
    assert sfs1.k_feature_names_ == (
        "sepal width",
        "petal width",
    ), sfs1.k_feature_names_
    assert (150, 2) == sfs1.transform(df).shape

    # now, test with different `feature_groups`
    sfs1 = SFS(
        lr,
        k_features=2,  # this is num of selected groups to form selected features
        forward=True,
        floating=False,
        scoring="accuracy",
        feature_groups=[
            ["petal length", "petal width"],
            ["sepal length"],
            ["sepal width"],
        ],
        cv=0,
        verbose=0,
        n_jobs=1,
    )

    sfs1 = sfs1.fit(df, y)
    # the selected fetures are sorted according their corresponding indices
    assert sfs1.k_feature_names_ == (
        "sepal width",
        "petal length",
        "petal width",
    ), sfs1.k_feature_names_


def test_check_pandas_dataframe_with_feature_groups_and_fixed_features():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=123)

    df = pd.DataFrame(
        X, columns=["sepal length", "sepal width", "petal length", "petal width"]
    )
    sfs1 = SFS(
        lr,
        k_features=2,
        forward=True,
        floating=False,
        scoring="accuracy",
        feature_groups=[
            ["petal length", "petal width"],
            ["sepal length"],
            ["sepal width"],
        ],
        fixed_features=["sepal length", "petal length", "petal width"],
        cv=0,
        verbose=0,
        n_jobs=1,
    )

    sfs1 = sfs1.fit(df, y)
    assert sfs1.k_feature_names_ == (
        "sepal length",
        "petal length",
        "petal width",
    ), sfs1.k_feature_names_


def test_check_feature_groups():
    iris = load_iris()
    X = iris.data
    y = iris.target
    lr = SoftmaxRegression(random_seed=123)
    sfs1 = SFS(
        lr,
        k_features=2,
        forward=True,
        floating=False,
        scoring="accuracy",
        feature_groups=[[2, 3], [0], [1]],
        fixed_features=[0, 2, 3],
        cv=0,
        verbose=0,
        n_jobs=1,
    )

    sfs1 = sfs1.fit(X, y)
    assert sfs1.k_feature_idx_ == (0, 2, 3), sfs1.k_feature_idx_



# mlxtend Machine Learning Library Extensions
# Author: Fatih Sen <fatih.sn2000@gmail.com>
#
# License: BSD 3 clause

import unittest

import numpy as np
import pandas as pd
from test_fpbase import (
    FPTestEdgeCases,
    FPTestErrors,
    FPTestEx1All,
    FPTestEx2All,
    FPTestEx3All,
    compare_dataframes,
)

from mlxtend.frequent_patterns import apriori, fpgrowth, hmine


class TestEdgeCases(unittest.TestCase, FPTestEdgeCases):
    def setUp(self):
        FPTestEdgeCases.setUp(self, hmine)


class TestErrors(unittest.TestCase, FPTestErrors):
    def setUp(self):
        FPTestErrors.setUp(self, hmine)


class TestHmine(unittest.TestCase, FPTestEx1All):
    def setUp(self):
        FPTestEx1All.setUp(self, hmine)


class TestHmineBoolInput(unittest.TestCase, FPTestEx1All):
    def setUp(self):
        one_ary = np.array(
            [
                [False, False, False, True, False, True, True, True, True, False, True],
                [False, False, True, True, False, True, False, True, True, False, True],
                [
                    True,
                    False,
                    False,
                    True,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    True,
                    False,
                    False,
                    True,
                    True,
                ],
                [
                    False,
                    True,
                    False,
                    True,
                    True,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                ],
            ]
        )
        FPTestEx1All.setUp(self, hmine, one_ary=one_ary)


class TestEx2(unittest.TestCase, FPTestEx2All):
    def setUp(self):
        FPTestEx2All.setUp(self, hmine)


class TestEx3(unittest.TestCase, FPTestEx3All):
    def setUp(self):
        FPTestEx3All.setUp(self, hmine)


class TestCorrect(unittest.TestCase):
    def setUp(self):
        self.one_ary = np.array(
            [
                [False, False, False, True, False, True, True, True, True, False, True],
                [False, False, True, True, False, True, False, True, True, False, True],
                [
                    True,
                    False,
                    False,
                    True,
                    False,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    False,
                    False,
                    False,
                    True,
                    True,
                    False,
                    False,
                    True,
                    True,
                ],
                [
                    False,
                    True,
                    False,
                    True,
                    True,
                    True,
                    False,
                    False,
                    True,
                    False,
                    False,
                ],
            ]
        )
        self.cols = [
            "Apple",
            "Corn",
            "Dill",
            "Eggs",
            "Ice cream",
            "Kidney Beans",
            "Milk",
            "Nutmeg",
            "Onion",
            "Unicorn",
            "Yogurt",
        ]
        self.df = pd.DataFrame(self.one_ary, columns=self.cols)

    def test_compare_correct(self):
        expect = pd.DataFrame(
            [
                [0.8, np.array([3])],
                [1.0, np.array([5])],
                [0.6, np.array([6])],
                [0.6, np.array([8])],
                [0.6, np.array([10])],
                [0.8, np.array([3, 5])],
                [0.6, np.array([3, 8])],
                [0.6, np.array([5, 6])],
                [0.6, np.array([5, 8])],
                [0.6, np.array([5, 10])],
                [0.6, np.array([3, 5, 8])],
            ],
            columns=["support", "itemsets"],
        )
        algorithms = {hmine: None, fpgrowth: None, apriori: None}
        for algo in algorithms.keys():
            self.setUp()
            res_df = algo(self.df, min_support=0.6)
            compare_dataframes(res_df, expect)
            algorithms[algo] = res_df

        compare_dataframes(algorithms[hmine], algorithms[fpgrowth])
        compare_dataframes(algorithms[hmine], algorithms[apriori])
        compare_dataframes(algorithms[fpgrowth], algorithms[apriori])



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

from mlxtend._base import _MultiLayer


class BlankClassifier(_MultiLayer):
    def __init__(self):
        pass


def test_layermapping_shape():
    mlp = BlankClassifier()
    w, b = mlp._layermapping(n_features=10, n_classes=11, hidden_layers=[8, 7, 6])

    expect_b = {
        "1": [[8], "n_hidden_1"],
        "2": [[7], "n_hidden_2"],
        "3": [[6], "n_hidden_3"],
        "out": [[11], "n_classes"],
    }

    expect_w = {
        "1": [[10, 8], "n_features, n_hidden_1"],
        "2": [[8, 7], "n_hidden_1, n_hidden_2"],
        "3": [[7, 6], "n_hidden_2, n_hidden_3"],
        "out": [[6, 11], "n_hidden_3, n_classes"],
    }

    assert expect_b == b, b
    assert expect_w == w, w


def test_init_from_layermapping():
    mlp = BlankClassifier()
    wm, bm = mlp._layermapping(n_features=5, n_classes=4, hidden_layers=[3, 2])
    w, b = mlp._init_params_from_layermapping(
        weight_maps=wm, bias_maps=bm, random_seed=1
    )

    assert len(w.keys()) == 3
    assert len(b.keys()) == 3
    assert set(w.keys()) == set(["1", "2", "out"])
    assert set(b.keys()) == set(["1", "2", "out"])

    assert w["1"].shape == (5, 3)
    assert w["2"].shape == (3, 2)
    assert w["out"].shape == (2, 4)

    assert b["1"].shape == (3,)
    assert b["2"].shape == (2,)
    assert b["out"].shape == (4,)


def test_init_from_layermapping_values():
    mlp = BlankClassifier()
    wm, bm = mlp._layermapping(n_features=5, n_classes=5, hidden_layers=[5])

    w, b = mlp._init_params_from_layermapping(
        weight_maps=wm, bias_maps=bm, random_seed=1
    )
    assert round(w["1"][0, 0], 8) == 0.01624345, w["1"][0, 0]
    assert round(w["out"][0, 0], 8) == -0.00683728, w["out"][0, 0]



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# A function for scoring predictions.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np

from mlxtend.evaluate.accuracy import accuracy_score
from mlxtend.evaluate.confusion_matrix import confusion_matrix


def _error(true, pred):
    return 1.0 - accuracy_score(true, pred, method="standard")


def _macro(true, pred, func, unique_labels):
    scores = []
    for lab in unique_labels:
        scores.append(func(np.where(true != lab, 1, 0), np.where(pred != lab, 1, 0)))
    return float(sum(scores)) / len(scores)


def scoring(
    y_target, y_predicted, metric="error", positive_label=1, unique_labels="auto"
):
    """Compute a scoring metric for supervised learning.

    Parameters
    ------------
    y_target : array-like, shape=[n_values]
        True class labels or target values.
    y_predicted : array-like, shape=[n_values]
        Predicted class labels or target values.
    metric : str (default: 'error')
        Performance metric:
        'accuracy': (TP + TN)/(FP + FN + TP + TN) = 1-ERR\n
        'average per-class accuracy': Average per-class accuracy\n
        'average per-class error':  Average per-class error\n
        'balanced per-class accuracy': Average per-class accuracy\n
        'balanced per-class error':  Average per-class error\n
        'error': (TP + TN)/(FP+ FN + TP + TN) = 1-ACC\n
        'false_positive_rate': FP/N = FP/(FP + TN)\n
        'true_positive_rate': TP/P = TP/(FN + TP)\n
        'true_negative_rate': TN/N = TN/(FP + TN)\n
        'precision': TP/(TP + FP)\n
        'recall': equal to 'true_positive_rate'\n
        'sensitivity': equal to 'true_positive_rate' or 'recall'\n
        'specificity': equal to 'true_negative_rate'\n
        'f1': 2 * (PRE * REC)/(PRE + REC)\n
        'matthews_corr_coef':  (TP*TN - FP*FN)
           / (sqrt{(TP + FP)( TP + FN )( TN + FP )( TN + FN )})\n
        Where:
        [TP: True positives, TN = True negatives,\n
         TN: True negatives, FN = False negatives]\n
    positive_label : int (default: 1)
        Label of the positive class for binary classification
        metrics.
    unique_labels : str or array-like (default: 'auto')
        If 'auto', deduces the unique class labels from
        y_target

    Returns
    ------------
    score : float

    Examples
    -----------
    For usage examples, please see
    https://rasbt.github.io/mlxtend/user_guide/evaluate/scoring/

    """
    implemented = {
        "error",
        "accuracy",
        "average per-class accuracy",
        "average per-class error",
        "balanced accuracy",
        "false_positive_rate",
        "true_positive_rate",
        "true_negative_rate",
        "precision",
        "recall",
        "sensitivity",
        "specificity",
        "matthews_corr_coef",
        "f1",
    }

    if metric not in implemented:
        raise AttributeError("`metric` not in %s" % implemented)

    if len(y_target) != len(y_predicted):
        raise AttributeError(
            "`y_target` and `y_predicted`" " don't have the same number of elements."
        )

    if unique_labels == "auto":
        unique_labels = np.unique(y_target)

    if not isinstance(y_target, np.ndarray):
        targ_tmp = np.asarray(y_target)
    else:
        targ_tmp = y_target
    if not isinstance(y_predicted, np.ndarray):
        pred_tmp = np.asarray(y_predicted)
    else:
        pred_tmp = y_predicted

    # multi-class metrics
    if metric == "accuracy":
        res = accuracy_score(targ_tmp, pred_tmp, method="standard")
    elif metric == "error":
        res = _error(targ_tmp, pred_tmp)
    elif metric == "average per-class accuracy":
        res = accuracy_score(targ_tmp, pred_tmp, method="average")
    elif metric == "average per-class error":
        res = _macro(targ_tmp, pred_tmp, func=_error, unique_labels=unique_labels)
    elif metric == "balanced accuracy":
        res = accuracy_score(targ_tmp, pred_tmp, method="balanced")

    # binary classification metrics
    else:
        if len(unique_labels) > 2 or len(np.unique(pred_tmp)) > 2:
            raise AttributeError(
                "Metrics precision, "
                "recall, and f1 only support binary"
                " class labels"
            )

        # `binary=True` makes sure
        # that positive label is 1 and negative label is 0
        cm = confusion_matrix(
            y_target=targ_tmp,
            y_predicted=pred_tmp,
            binary=True,
            positive_label=positive_label,
        )
        tp = cm[-1, -1]
        fp = cm[0, -1]
        tn = cm[0, 0]
        fn = cm[-1, 0]

        if metric == "false_positive_rate":
            res = float(fp) / (fp + tn)
        elif metric in {"true_positive_rate", "recall", "sensitivity"}:
            res = float(tp) / (fn + tp)
        elif metric in {"true_negative_rate", "specificity"}:
            res = float(tn) / (fp + tn)
        elif metric == "precision":
            res = float(tp) / (tp + fp)
        elif metric == "f1":
            pre = float(tp) / (tp + fp)
            rec = float(tp) / (fn + tp)
            res = 2.0 * (pre * rec) / (pre + rec)
        elif metric == "matthews_corr_coef":
            res = float(tp * tn - fp * fn)
            res = res / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))

    return res



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause


import numpy as np
import pandas as pd

from mlxtend.preprocessing import minmax_scaling


def test_minmax_scaling_arrayerror():
    try:
        ary = [[1, 2], [3, 4]]
        minmax_scaling(ary, [1, "s2"])
    except AttributeError:
        pass
    else:
        raise AssertionError


def test_pandas_minmax_scaling():
    s1 = pd.Series([1, 2, 3, 4, 5, 6], index=(range(6)))
    s2 = pd.Series([10, 9, 8, 7, 6, 5], index=(range(6)))
    df = pd.DataFrame(s1, columns=["s1"])
    df["s2"] = s2

    df_out1 = minmax_scaling(df, ["s1", "s2"], min_val=0, max_val=1)
    df_out2 = minmax_scaling(df, ["s1", "s2"], min_val=50, max_val=100)

    ary_out1 = np.array(
        [[0.0, 1.0], [0.2, 0.8], [0.4, 0.6], [0.6, 0.4], [0.8, 0.2], [1.0, 0.0]]
    )

    ary_out2 = np.array(
        [
            [50.0, 100.0],
            [60.0, 90.0],
            [70.0, 80.0],
            [80.0, 70.0],
            [90.0, 60.0],
            [100.0, 50.0],
        ]
    )

    np.testing.assert_allclose(df_out1.values, ary_out1, rtol=1e-03)
    assert (df_out2.values == ary_out2).all()


def test_numpy_minmax_scaling():
    ary = np.array([[1, 10], [2, 9], [3, 8], [4, 7], [5, 6], [6, 5]])

    df_out1 = minmax_scaling(ary, [0, 1], min_val=0, max_val=1)
    df_out2 = minmax_scaling(ary, [0, 1], min_val=50, max_val=100)

    ary_out1 = np.array(
        [[0.0, 1.0], [0.2, 0.8], [0.4, 0.6], [0.6, 0.4], [0.8, 0.2], [1.0, 0.0]]
    )

    ary_out2 = np.array(
        [
            [50.0, 100.0],
            [60.0, 90.0],
            [70.0, 80.0],
            [80.0, 70.0],
            [90.0, 60.0],
            [100.0, 50.0],
        ]
    )

    np.testing.assert_allclose(df_out1, ary_out1, rtol=1e-03)
    assert (df_out2 == ary_out2).all()



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from packaging.version import Version
from scipy import sparse
from sklearn import __version__ as sklearn_version
from sklearn.base import clone
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression, Ridge
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR

from mlxtend.externals.estimator_checks import NotFittedError
from mlxtend.regressor import StackingRegressor
from mlxtend.utils import assert_raises

# Generating a sample dataset
np.random.seed(1)
X1 = np.sort(5 * np.random.rand(40, 1), axis=0)
X2 = np.sort(5 * np.random.rand(40, 2), axis=0)
y = np.sin(X1).ravel()
y[::5] += 3 * (0.5 - np.random.rand(8))
y2 = np.sin(X2)
w = np.random.random(40)


def test_different_models():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stregr = StackingRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf)
    stregr.fit(X1, y).predict(X1)
    mse = 0.21
    got = np.mean((stregr.predict(X1) - y) ** 2)
    assert round(got, 2) == mse


def test_multivariate():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stregr = StackingRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf)
    stregr.fit(X2, y).predict(X2)
    mse = 0.22
    got = np.mean((stregr.predict(X2) - y) ** 2)
    assert round(got, 2) == mse


def test_multivariate_class():
    lr = LinearRegression()
    ridge = Ridge(random_state=1)
    meta = LinearRegression()
    stregr = StackingRegressor(
        regressors=[lr, ridge], meta_regressor=meta, multi_output=True
    )
    stregr.fit(X2, y2).predict(X2)
    mse = 0.12
    got = np.mean((stregr.predict(X2) - y2) ** 2.0)
    assert round(got, 2) == mse, got


def test_sample_weight():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stregr = StackingRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf)
    pred1 = stregr.fit(X1, y, sample_weight=w).predict(X1)
    mse = 0.22
    got = np.mean((stregr.predict(X1) - y) ** 2)
    assert round(got, 2) == mse
    # make sure that this is not equivalent to the model with no weight
    pred2 = stregr.fit(X1, y).predict(X1)
    maxdiff = np.max(np.abs(pred1 - pred2))
    assert maxdiff > 1e-3, "max diff is %.4f" % maxdiff


def test_weight_ones():
    # sample weight of ones should produce equivalent outcome as no weight
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stregr = StackingRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf)
    pred1 = stregr.fit(X1, y).predict(X1)
    pred2 = stregr.fit(X1, y, sample_weight=np.ones(40)).predict(X1)
    maxdiff = np.max(np.abs(pred1 - pred2))
    assert maxdiff < 1e-3, "max diff is %.4f" % maxdiff


def test_weight_unsupported_regressor():
    # including regressor that does not support
    # sample_weight should raise error
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    knn = KNeighborsRegressor()
    stregr = StackingRegressor(
        regressors=[svr_lin, lr, ridge, knn], meta_regressor=svr_rbf
    )

    with pytest.raises(TypeError):
        stregr.fit(X1, y, sample_weight=w).predict(X1)


def test_weight_unsupported_meta():
    # meta regressor with no support for
    # sample_weight should raise error
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    knn = KNeighborsRegressor()
    stregr = StackingRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=knn)

    with pytest.raises(TypeError):
        stregr.fit(X1, y, sample_weight=w).predict(X1)


def test_weight_unsupported_with_no_weight():
    # pass no weight to regressors with no weight support
    # should not be a problem
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    knn = KNeighborsRegressor()
    stregr = StackingRegressor(
        regressors=[svr_lin, lr, ridge, knn], meta_regressor=svr_rbf
    )
    stregr.fit(X1, y).predict(X1)

    stregr = StackingRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=knn)
    stregr.fit(X1, y).predict(X1)


def test_gridsearch():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stregr = StackingRegressor(regressors=[svr_lin, lr, ridge], meta_regressor=svr_rbf)

    params = {
        "ridge__alpha": [0.01, 1.0],
        "svr__C": [0.01, 1.0],
        "meta_regressor__C": [0.01, 1.0],
    }

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(
            estimator=stregr, param_grid=params, cv=5, iid=False, refit=True, verbose=0
        )
    else:
        grid = GridSearchCV(
            estimator=stregr, param_grid=params, cv=5, refit=True, verbose=0
        )
    grid = grid.fit(X1, y)
    best = 0.1
    got = round(grid.best_score_, 2)
    assert best == got


def test_gridsearch_numerate_regr():
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stregr = StackingRegressor(
        regressors=[svr_lin, ridge, ridge], meta_regressor=svr_rbf
    )

    params = {
        "ridge-1__alpha": [0.01, 1.0],
        "ridge-2__alpha": [0.01, 1.0],
        "svr__C": [0.01, 1.0],
        "meta_regressor__C": [0.01, 1.0],
    }

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(
            estimator=stregr, param_grid=params, cv=5, iid=False, refit=True, verbose=0
        )
    else:
        grid = GridSearchCV(
            estimator=stregr, param_grid=params, cv=5, refit=True, verbose=0
        )
    grid = grid.fit(X1, y)
    best = 0.1
    got = round(grid.best_score_, 2)
    assert best == got


def test_get_coeff():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(regressors=[svr_lin, lr], meta_regressor=ridge)
    stregr.fit(X1, y)
    got = stregr.coef_
    expect = np.array([0.4874216, 0.45518317])
    assert_almost_equal(got, expect)


def test_get_intercept():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(regressors=[svr_lin, lr], meta_regressor=ridge)
    stregr.fit(X1, y)
    got = stregr.intercept_
    expect = 0.02
    assert round(got, 2) == expect


# ValueError was changed to AttributeError in sklearn >= 0.19
def test_get_coeff_fail():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(regressors=[ridge, lr], meta_regressor=svr_rbf)

    with pytest.raises(AttributeError):
        stregr = stregr.fit(X1, y)
        r = stregr.coef_
        assert r


def test_get_params():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(regressors=[ridge, lr], meta_regressor=svr_rbf)

    got = sorted(list({s.split("__")[0] for s in stregr.get_params().keys()}))
    expect = [
        "linearregression",
        "meta_regressor",
        "multi_output",
        "refit",
        "regressors",
        "ridge",
        "store_train_meta_features",
        "use_features_in_secondary",
        "verbose",
    ]
    assert got == expect, got


def test_regressor_gridsearch():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(regressors=[lr], meta_regressor=svr_rbf)

    params = {"regressors": [[lr], [lr, ridge]]}

    if Version(sklearn_version) < Version("0.24.1"):
        grid = GridSearchCV(
            estimator=stregr, param_grid=params, cv=5, iid=False, refit=True
        )
    else:
        grid = GridSearchCV(estimator=stregr, param_grid=params, cv=5, refit=True)
    grid.fit(X1, y)

    assert len(grid.best_params_["regressors"]) == 2


def test_predict_meta_features():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(regressors=[lr, ridge], meta_regressor=svr_rbf)
    X_train, X_test, y_train, y_test = train_test_split(X2, y, test_size=0.3)
    stregr.fit(X_train, y_train)
    test_meta_features = stregr.predict(X_test)
    assert test_meta_features.shape[0] == X_test.shape[0]


def test_train_meta_features_():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(
        regressors=[lr, ridge], meta_regressor=svr_rbf, store_train_meta_features=True
    )
    X_train, X_test, y_train, y_test = train_test_split(X2, y, test_size=0.3)
    stregr.fit(X_train, y_train)
    train_meta_features = stregr.train_meta_features_
    assert train_meta_features.shape[0] == X_train.shape[0]


def test_not_fitted_predict():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(
        regressors=[lr, ridge], meta_regressor=svr_rbf, store_train_meta_features=True
    )
    X_train, X_test, y_train, y_test = train_test_split(X2, y, test_size=0.3)

    expect = (
        "This StackingRegressor instance is not fitted yet. Call "
        "'fit' with appropriate arguments before using this method."
    )

    assert_raises(NotFittedError, expect, stregr.predict, X_train)

    assert_raises(NotFittedError, expect, stregr.predict_meta_features, X_train)


def test_clone():
    lr = LinearRegression()
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(
        regressors=[lr, ridge], meta_regressor=svr_rbf, store_train_meta_features=True
    )
    clone(stregr)


def test_features_in_secondary():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    rf = RandomForestRegressor(n_estimators=10, random_state=2)
    ridge = Ridge(random_state=0)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingRegressor(
        regressors=[svr_lin, lr, ridge, rf],
        meta_regressor=svr_rbf,
        use_features_in_secondary=True,
    )

    stack.fit(X1, y).predict(X1)
    mse = 0.14
    got = np.mean((stack.predict(X1) - y) ** 2)
    print(got)
    assert round(got, 2) == mse

    stack = StackingRegressor(
        regressors=[svr_lin, lr, ridge, rf],
        meta_regressor=svr_rbf,
        use_features_in_secondary=False,
    )

    # dense
    stack.fit(X1, y).predict(X1)
    mse = 0.12
    got = np.mean((stack.predict(X1) - y) ** 2)
    print(got)
    assert round(got, 2) == mse


def test_predictions_from_sparse_matrix():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    ridge = Ridge(random_state=1)
    stregr = StackingRegressor(regressors=[svr_lin, lr], meta_regressor=ridge)

    # dense
    stregr.fit(X1, y)
    print(stregr.score(X1, y))
    assert round(stregr.score(X1, y), 2) == 0.61

    # sparse
    stregr.fit(sparse.csr_matrix(X1), y)
    print(stregr.score(X1, y))
    assert round(stregr.score(X1, y), 2) == 0.61


def test_sparse_matrix_inputs_and_features_in_secondary():
    lr = LinearRegression()
    svr_lin = SVR(kernel="linear", gamma="auto")
    rf = RandomForestRegressor(n_estimators=10, random_state=2)
    ridge = Ridge(random_state=0)
    svr_rbf = SVR(kernel="rbf", gamma="auto")
    stack = StackingRegressor(
        regressors=[svr_lin, lr, ridge, rf],
        meta_regressor=svr_rbf,
        use_features_in_secondary=True,
    )

    # dense
    stack.fit(X1, y).predict(X1)
    mse = 0.14
    got = np.mean((stack.predict(X1) - y) ** 2)
    assert round(got, 2) == mse

    # sparse
    stack.fit(sparse.csr_matrix(X1), y)
    mse = 0.14
    got = np.mean((stack.predict(sparse.csr_matrix(X1)) - y) ** 2)
    assert round(got, 2) == mse



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# Object for selecting a dataset column in scikit-learn pipelines.
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: BSD 3 clause

import numpy as np
import pandas as pd
from packaging.version import Version
from sklearn import __version__ as sklearn_version
from sklearn import datasets
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import make_pipeline

from mlxtend.feature_selection import ColumnSelector


def test_ColumnSelector():
    X1_in = np.ones((4, 8))
    X1_out = ColumnSelector(cols=(1, 3)).transform(X1_in)
    assert X1_out.shape == (4, 2)


def test_ColumnSelector_drop_axis():
    X1_in = np.ones((4, 8))
    X1_out = ColumnSelector(cols=1, drop_axis=True).transform(X1_in)
    assert X1_out.shape == (4,)

    X1_out = ColumnSelector(cols=(1,), drop_axis=True).transform(X1_in)
    assert X1_out.shape == (4,)

    X1_out = ColumnSelector(cols=1).transform(X1_in)
    assert X1_out.shape == (4, 1)

    X1_out = ColumnSelector(cols=(1,)).transform(X1_in)
    assert X1_out.shape == (4, 1)


def test_ColumnSelector_in_gridsearch():
    iris = datasets.load_iris()
    X, y = iris.data, iris.target
    pipe = make_pipeline(
        ColumnSelector(), LogisticRegression(multi_class="ovr", solver="liblinear")
    )
    grid = {
        "columnselector__cols": [[1, 2], [1, 2, 3], 0, [1]],
        "logisticregression__C": [0.1, 1.0, 10.0],
    }

    if Version(sklearn_version) < Version("0.24.1"):
        gsearch1 = GridSearchCV(
            estimator=pipe, param_grid=grid, iid=False, cv=5, n_jobs=1, refit=False
        )
    else:
        gsearch1 = GridSearchCV(
            estimator=pipe, param_grid=grid, cv=5, n_jobs=1, refit=False
        )

    gsearch1.fit(X, y)
    assert gsearch1.best_params_["columnselector__cols"] == [1, 2, 3]


def test_ColumnSelector_with_dataframe():
    iris = datasets.load_iris()
    df_in = pd.DataFrame(iris.data, columns=iris.feature_names)
    df_out = ColumnSelector(cols=("sepal length (cm)", "sepal width (cm)")).transform(
        df_in
    )
    assert df_out.shape == (150, 2)


def test_ColumnSelector_with_dataframe_and_int_columns():
    iris = datasets.load_iris()
    df_in = pd.DataFrame(iris.data, columns=iris.feature_names)
    df_out_str = ColumnSelector(
        cols=("petal length (cm)", "petal width (cm)")
    ).transform(df_in)
    df_out_int = ColumnSelector(cols=(2, 3)).transform(df_in)

    np.testing.assert_array_equal(df_out_str[:, 0], df_out_int[:, 0])
    np.testing.assert_array_equal(df_out_str[:, 1], df_out_int[:, 1])


def test_ColumnSelector_with_dataframe_drop_axis():
    iris = datasets.load_iris()
    df_in = pd.DataFrame(iris.data, columns=iris.feature_names)
    X1_out = ColumnSelector(cols=("petal length (cm)",), drop_axis=True).transform(
        df_in
    )
    assert X1_out.shape == (150,)

    X1_out = ColumnSelector(cols=("petal length (cm)",), drop_axis=True).transform(
        df_in
    )
    assert X1_out.shape == (150,)

    X1_out = ColumnSelector(cols="petal length (cm)").transform(df_in)
    assert X1_out.shape == (150, 1)

    X1_out = ColumnSelector(cols=("petal length (cm)",)).transform(df_in)
    assert X1_out.shape == (150, 1)


def test_ColumnSelector_with_dataframe_in_gridsearch():
    iris = datasets.load_iris()
    X = pd.DataFrame(iris.data, columns=iris.feature_names)
    y = iris.target
    pipe = make_pipeline(ColumnSelector(), LogisticRegression())
    grid = {
        "columnselector__cols": [
            ["petal length (cm)", "petal width (cm)"],
            ["sepal length (cm)", "sepal width (cm)", "petal width (cm)"],
        ],
    }

    gsearch1 = GridSearchCV(
        estimator=pipe,
        param_grid=grid,
        cv=5,
        n_jobs=1,
        scoring="accuracy",
        refit=False,
    )

    gsearch1.fit(X, y)
    assert gsearch1.best_params_["columnselector__cols"] == [
        "petal length (cm)",
        "petal width (cm)",
    ]



# Sebastian Raschka 2014-2024
# mlxtend Machine Learning Library Extensions
#
# License: BSD 3 clause
import numpy as np
from numpy.testing import assert_almost_equal

from mlxtend.evaluate import accuracy_score


def test_multiclass_binary():
    y_targ = [0, 0, 0, 1, 1, 1, 2, 2, 2]
    y_pred = [1, 0, 0, 0, 1, 2, 0, 2, 2]

    y_1 = accuracy_score(y_targ, y_pred, method="binary", pos_label=1)
    y_2 = accuracy_score(y_targ, y_pred, method="binary", pos_label=2)
    y_3 = accuracy_score(y_targ, y_pred, method="binary", pos_label=1, normalize=False)

    assert_almost_equal(y_2, float(7) / 9, decimal=4)
    assert_almost_equal(y_1, float(6) / 9, decimal=4)
    assert y_3 == 6


def test_standard():
    y_targ = [0, 0, 0, 1, 1, 1]
    y_pred = [0, 1, 1, 0, 1, 1]

    y = accuracy_score(y_targ, y_pred, method="standard")
    y_1 = accuracy_score(y_targ, y_pred, method="binary", normalize=False)
    assert_almost_equal(y, float(3) / 6, decimal=4)
    assert y_1 == 3


def test_balanced_multiclass():
    y_targ = np.array([0, 0, 0, 1, 1, 1, 1, 1, 2, 2])
    y_pred = np.array([0, 1, 1, 0, 1, 1, 2, 2, 2, 2])

    y = accuracy_score(y_targ, y_pred, method="balanced")
    assert_almost_equal(y, 0.578, decimal=3)


def test_balanced_binary():
    y_targ = np.array([0, 0, 0, 1, 1, 1, 1, 1, 1, 0])
    y_pred = np.array([0, 1, 1, 0, 1, 1, 1, 1, 1, 1])

    y = accuracy_score(y_targ, y_pred, method="balanced")
    assert_almost_equal(y, 0.542, decimal=3)


def test_average():
    y_targ = np.array([0, 0, 0, 1, 1, 1, 1, 1, 2, 2])
    y_pred = np.array([0, 1, 1, 0, 1, 1, 2, 2, 2, 2])

    y = accuracy_score(y_targ, y_pred, method="average")
    assert_almost_equal(y, float(2) / 3, decimal=4)
